import inspect
from typing import Any, Callable, Type, Generic, Optional, List
from .constants import P, R


class BaseWire(Generic[P, R]):
    def __init__(self, func: Callable, instance: Any, owner: Type, name: str, plugins: Optional[List] = None):
        self._func = func
        self._instance = instance
        self._owner = owner
        self._name = name
        self.plugins = plugins or []
        self._early_response = None
    
    def _invoke_flexible(self, func: Callable, *args: Any, **kwargs: Any) -> Any:
        if self._instance is not None:
            return func(self._instance, *args, **kwargs)
        return func(*args, **kwargs)
    
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        self._early_response = None
        for p in self.plugins:
            # ПРОВЕРКА: BaseWire не терпит асинхронных плагинов
            if inspect.iscoroutinefunction(getattr(p, 'before', None)):
                raise TypeError(f"Plugin {p.__class__.__name__} is async, but wire {self._name} is sync!")
            
            if hasattr(p, 'before'):
                args, kwargs = p.before(self, args, kwargs)
            
            if self._early_response is not None:
                return self._early_response
        
        result = self._invoke_flexible(self._func, *args, **kwargs)
        
        for p in reversed(self.plugins):
            if inspect.iscoroutinefunction(getattr(p, 'after', None)):
                raise TypeError(f"Plugin {p.__class__.__name__} is async, but wire {self._name} is sync!")
            
            if hasattr(p, 'after'):
                result = p.after(self, result)
        return result


class AsyncWire(BaseWire[P, R]):
    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._execute_async(*args, **kwargs)
    
    async def _execute_async(self, *args: Any, **kwargs: Any) -> Any:
        self._early_response = None
        for p in self.plugins:
            # AsyncWire работает со всеми: и sync, и async
            before_method = getattr(p, 'before', None)
            if before_method:
                if inspect.iscoroutinefunction(before_method):
                    args, kwargs = await before_method(self, args, kwargs)
                else:
                    args, kwargs = before_method(self, args, kwargs)
            
            if self._early_response is not None:
                return self._early_response
        
        res = self._invoke_flexible(self._func, *args, **kwargs)
        result = await res if inspect.isawaitable(res) else res
        
        for p in reversed(self.plugins):
            after_method = getattr(p, 'after', None)
            if after_method:
                if inspect.iscoroutinefunction(after_method):
                    result = await after_method(self, result)
                else:
                    result = after_method(self, result)
        return result
    
    def __await__(self):
        return self._execute_async().__await__()


from typing import Protocol, Any, runtime_checkable
from .constants import P, R
from .wire import BaseWire, AsyncWire


@runtime_checkable
class SyncPlugin(Protocol):
    def before(self, wire: BaseWire[P, R], args: tuple, kwargs: dict) -> tuple[tuple, dict]: return args, kwargs
    
    def after(self, wire: BaseWire[P, R], result: Any) -> Any: return result


@runtime_checkable
class AsyncPlugin(Protocol):
    async def before(self, wire: AsyncWire[P, R], args: tuple, kwargs: dict) -> tuple[tuple, dict]: return args, kwargs
    
    async def after(self, wire: AsyncWire[P, R], result: Any) -> Any: return result



import weakref
import functools
import inspect

from .wire import BaseWire, AsyncWire


def support_ropes(cls):
    if not hasattr(cls, "__slots__"): return cls
    if "__weakref__" in cls.__slots__: return cls
    new_slots = tuple(cls.__slots__) + ("__weakref__",)
    namespace = {k: v for k, v in cls.__dict__.items()
                 if k not in ("__slots__", "__weakref__") and k not in cls.__slots__}
    return type(cls.__name__, cls.__bases__, {"__slots__": new_slots, **namespace})


class SmartRope:
    def __init__(self, wrapper_class=None, plugins: list = None):
        self.wire_class = wrapper_class
        self.plugins = plugins or []
        self.raw_func = None
        self.name = None
        self._wires = weakref.WeakKeyDictionary()
        self._global_wire = None
    
    def register(self, *types):
        def wrapper(func):
            target = next((p for p in self.plugins if hasattr(p, "register")), None)
            if not target:
                raise AttributeError("No plugin in this SmartRope supports .register()")
            target.register(types, func)
            return func
        
        return wrapper
    
    def __call__(self, *args, **kwargs):
        if self.raw_func is None:
            func = args[0]
            if isinstance(func, SmartRope):
                self.plugins = func.plugins + self.plugins
                if self.wire_class is None: self.wire_class = func.wire_class
                self.raw_func = func.raw_func
            else:
                self.raw_func = func
            if self.raw_func:
                functools.update_wrapper(self, self.raw_func)
            return self
        
        if self._global_wire is None:
            self._global_wire = self._get_wire(None, None)
        return self._global_wire(*args, **kwargs)
    
    def _get_wire(self, instance, owner):
        if instance is None and owner is None:
            if self._global_wire is None:
                self._global_wire = self._create_wire(None, None)
            return self._global_wire
        
        key = instance if instance is not None else owner
        if key not in self._wires:
            self._wires[key] = self._create_wire(instance, owner)
        return self._wires[key]
    
    def _create_wire(self, instance, owner):
        """Вынес логику создания в отдельный метод для чистоты."""
        obj = self.raw_func
        kind, func = 'method', obj
        if isinstance(obj, staticmethod):
            kind, func = 'static', obj.__func__
        elif isinstance(obj, classmethod):
            kind, func = 'class', obj.__func__
        elif isinstance(obj, property):
            kind, func = 'property', obj.fget
        
        actual_func = func
        actual_instance = instance
        if kind == 'class':
            actual_func = func.__get__(None, owner)
            actual_instance = None
        elif kind == 'static':
            actual_instance = None
        
        w_cls = self.wire_class or (AsyncWire if inspect.iscoroutinefunction(func) else BaseWire)
        
        return w_cls(
            func=actual_func, instance=actual_instance,
            owner=owner, name=self.name or getattr(func, "__name__", "unknown"),
            plugins=self.plugins
        )
    
    def __get__(self, instance, owner):
        if instance is None and not isinstance(self.raw_func, (classmethod, staticmethod)):
            return self
        return self._get_wire(instance, owner)
    
    def __set_name__(self, owner, name):
        self.name = name


from .rope import SmartRope


def rope_factory(plugin_cls, wrapper_class=None, **default_plugin_kwargs):
    def decorator(func=None, **call_kwargs):
        final_params = {**default_plugin_kwargs, **call_kwargs}
        
        def wrapper(f):
            p = plugin_cls(**final_params)
            return SmartRope(wrapper_class, plugins=[p])(f)
        
        return wrapper(func) if (func is not None and callable(func)) else wrapper
    
    return decorator



from typing import TypeVar, ParamSpec, Callable, Any, Type, Protocol

P = ParamSpec("P")
R = TypeVar("R")


class WireProtocol(Protocol[P, R]):
    _func: Callable
    _instance: Any
    _owner: Type
    _name: str
    
    def _invoke_flexible(self, func: Callable, *args: Any, **kwargs: Any) -> Any: ...
